"""Sphinx configuration file for Flow Gym documentation."""

import os
import sys
from dataclasses import asdict

sys.path.insert(0, os.path.abspath(".."))

from sphinxawesome_theme import ThemeOptions
from sphinxawesome_theme.postprocess import Icons

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Flow Gym"
copyright = "2025, Cristian Perez Jensen"
author = "Cristian Perez Jensen"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosummary",
    "sphinx.ext.mathjax",
    "myst_parser",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# Suppress warnings
suppress_warnings = ["ref.ref", "autosummary"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_title = project
html_theme = "sphinxawesome_theme"
html_last_updated_fmt = ""
html_use_index = False
html_domain_indices = False
html_copy_source = False
html_permalinks_icon = Icons.permalinks_icon
html_baseurl = "https://cristianpjensen.github.io/flowgym/"
html_context = {
    "mode": "production",
}

autosummary_generate = True
autosummary_format_signature = ""
add_module_names = False

# Autodoc settings
autodoc_default_options = {
    "inherited-members": False,
}

theme_options = ThemeOptions(
    show_prev_next=True,
    awesome_external_links=True,
    main_nav_links={"Docs": "/flowgym/index", "Changelog": "/flowgym/changelog/index"},
    extra_header_link_icons={
        "Repository on GitHub": {
            "link": "https://github.com/cristianpjensen/flowgym",
            "icon": (
                '<svg height="26px" style="margin-top:-2px;display:inline" '
                'viewBox="0 0 45 44" '
                'fill="currentColor" xmlns="http://www.w3.org/2000/svg">'
                '<path fill-rule="evenodd" clip-rule="evenodd" '
                'd="M22.477.927C10.485.927.76 10.65.76 22.647c0 9.596 6.223 17.736 '
                "14.853 20.608 1.087.2 1.483-.47 1.483-1.047 "
                "0-.516-.019-1.881-.03-3.693-6.04 "
                "1.312-7.315-2.912-7.315-2.912-.988-2.51-2.412-3.178-2.412-3.178-1.972-1.346.149-1.32.149-1.32 "  # noqa
                "2.18.154 3.327 2.24 3.327 2.24 1.937 3.318 5.084 2.36 6.321 "
                "1.803.197-1.403.759-2.36 "
                "1.379-2.903-4.823-.548-9.894-2.412-9.894-10.734 "
                "0-2.37.847-4.31 2.236-5.828-.224-.55-.969-2.759.214-5.748 0 0 "
                "1.822-.584 5.972 2.226 "
                "1.732-.482 3.59-.722 5.437-.732 1.845.01 3.703.25 5.437.732 "
                "4.147-2.81 5.967-2.226 "
                "5.967-2.226 1.185 2.99.44 5.198.217 5.748 1.392 1.517 2.232 3.457 "
                "2.232 5.828 0 "
                "8.344-5.078 10.18-9.916 10.717.779.67 1.474 1.996 1.474 4.021 0 "
                "2.904-.027 5.247-.027 "
                "5.96 0 .58.392 1.256 1.493 1.044C37.981 40.375 44.2 32.24 44.2 "
                '22.647c0-11.996-9.726-21.72-21.722-21.72" '
                'fill="currentColor"/></svg>'
            ),
        },
    },
)

html_theme_options = asdict(theme_options)

sitemap_show_lastmod = False
sitemap_url_scheme = "{link}"



"""flowgym package."""

from flowgym.base_models import BaseModel
from flowgym.environments import (
    EndpointEnvironment,
    Environment,
    EpsilonEnvironment,
    ScoreEnvironment,
    VelocityEnvironment,
)
from flowgym.make import make
from flowgym.registry import base_model_registry, reward_registry
from flowgym.rewards import Reward
from flowgym.schedulers import (
    ConstantNoiseSchedule,
    CosineScheduler,
    DiffusionScheduler,
    MemorylessNoiseSchedule,
    NoiseSchedule,
    OptimalTransportScheduler,
    Scheduler,
)
from flowgym.types import DataProtocol, DataType, FGTensor

__all__ = [
    "BaseModel",
    "ConstantNoiseSchedule",
    "CosineScheduler",
    "DataProtocol",
    "DataType",
    "DiffusionScheduler",
    "EndpointEnvironment",
    "Environment",
    "EpsilonEnvironment",
    "FGTensor",
    "MemorylessNoiseSchedule",
    "NoiseSchedule",
    "OptimalTransportScheduler",
    "Reward",
    "Scheduler",
    "ScoreEnvironment",
    "VelocityEnvironment",
    "base_model_registry",
    "make",
    "reward_registry",
]

try:
    from . import molecules

    HAS_MOLECULES = True
except ImportError:
    HAS_MOLECULES = False

try:
    from . import images

    HAS_IMAGES = True
except ImportError:
    HAS_IMAGES = False

if HAS_MOLECULES:
    __all__ += ["molecules"]

if HAS_IMAGES:
    __all__ += ["images"]



"""Reward function for protein-ligand docking tasks using AutoDock-GPU."""

import os
from typing import Any, Optional

import torch
from meeko import MoleculePreparation, PDBQTWriterLegacy
from rdkit import Chem

from flowgym import reward_registry
from flowgym.molecules import BatchedMoleculeReward
from flowgym.utils import temporary_workdir


@reward_registry.register("molecules/protein_docking")
class ProteinDockingReward(BatchedMoleculeReward):
    """Reward function for protein-ligand docking tasks.

    Requires AutoDock-GPU to be installed, a prepared protein, and a pre-computed grid.
    Assuming you have a protein with a ligand included, you can split them using the following gist:
    https://gist.github.com/cristianpjensen/16024f1d576423ec510acdcabb2b74bd. Once they have been
    split, you need to prepare the protein and compute the grid maps:
    ```
    mk_prepare_receptor.py --read_pdb <protein>_protein.pdb -o <protein> -p -g \
        --box_enveloping <protein>_ligand.pdb --padding 5
    autogrid4 -p <protein>.gpf -l <protein>.glg
    ```
    Then specify the directory with all these files as `protein_dir`.

    Parameters
    ----------
    protein_dir : str
        Directory containing the protein structure files.

    atom_type_map : list[str], default=["C", "H", "N", "O", "F", "P", "S", "Cl", "Br", "I"]
        List of atom types corresponding to the node features in the input graph. The index of each
        atom type in the list corresponds to its integer representation in the node features.

    validate : bool, default=True
        Whether to validate molecules using RDKit's sanitization.

    check_fragmented : bool, default=True
        Whether to allow fragmented molecules (i.e., molecules with multiple disconnected
        components).

    relax : bool, default=False
        Whether to relax the geometry of molecules using MMFF optimization. If using `relax`, you
        probably should also use `validate`, since otherwise there is a high chance that your
        program will crash.
    """

    def __init__(
        self,
        protein_dir: str,
        atom_type_map: Optional[list[str]] = None,
        validate: bool = True,
        check_fragmented: bool = True,
        relax: bool = False,
    ):
        super().__init__(
            atom_type_map=atom_type_map,
            validate=validate,
            check_fragmented=check_fragmented,
            relax=relax,
        )
        self.protein_dir = os.path.abspath(protein_dir)

    def compute_rewards(
        self,
        mols: list[Chem.Mol],
        indices: list[int],
        **kwargs: Any,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Compute the docking score of the molecules to the specified protein.

        Parameters
        ----------
        mol : list[Chem.Mol]
            The molecules to compute the reward for.

        Returns
        -------
        rewards : torch.Tensor, shape (len(mols),)
            The computed rewards.
        """
        out = torch.zeros(len(mols))
        valids = torch.zeros(len(mols), dtype=torch.bool)
        file_list = []

        with temporary_workdir():
            os.mkdir("ligands")

            # 1. Prepare all ligands for docking and keep track of all files so we can batch the
            # autodock call using a file list
            mk_prep = MoleculePreparation()
            for i, mol in enumerate(mols):
                try:
                    molsetup_list = mk_prep(mol)
                    molsetup = molsetup_list[0]
                    pdbqt_string, success, _ = PDBQTWriterLegacy.write_string(molsetup)
                    if not success:
                        continue
                except Exception:
                    continue

                fname = f"ligand_{i:05d}.pdbqt"
                with open(os.path.join("ligands", fname), "w") as f:
                    f.write(pdbqt_string)

                file_list.append(fname)

            # 2. Write file list to disk
            with open("ligands/file_list.txt", "w") as f:
                f.write(str(os.path.join(self.protein_dir, "protein.maps.fld")) + "\n")
                for fname in file_list:
                    f.write(fname + "\n")

            # 3. Run autodock GPU in batch mode
            os.system("autodock_gpu_128wi -B ligands/file_list.txt > /dev/null 2>&1")

            # 4. Extract docking scores from output files
            for i in range(len(mols)):
                fname = f"ligands/ligand_{i:05d}.dlg"
                if not os.path.exists(fname):
                    valids[i] = False
                    continue

                with open(fname, "r") as f:
                    for line in f:
                        if line.strip().endswith("RANKING"):
                            parts = line.split()
                            binding_affinity = float(parts[3])
                            out[i] = -binding_affinity
                            valids[i] = True

        return out, valids



"""Common noise schedules for flow matching and diffusion models."""

import torch

from flowgym.types import DataType

from .base import NoiseSchedule


class ConstantNoiseSchedule(NoiseSchedule[DataType]):
    """Constant noise schedule with fixed sigma.

    Parameters
    ----------
    sigma : float
        Constant noise level.
    """

    def __init__(self, sigma: float):
        self.sigma = sigma

    def noise(self, t: torch.Tensor) -> torch.Tensor:
        """Compute the noise level at time t."""
        return self.sigma * torch.ones_like(t).unsqueeze(-1)

    def __call__(self, x: DataType, t: torch.Tensor) -> DataType:
        """Constant noise schedule."""
        return self.sigma * x.ones_like()



"""Utility functions for molecular reward calculations."""

import os
from typing import Optional

from rdkit import Chem

from flowgym.utils import temporary_workdir


def validate_mol(mol: Chem.Mol) -> Optional[Chem.Mol]:
    """Validate molecules according to chemical validity.

    Source: https://arxiv.org/abs/2505.00518

    Parameters
    ----------
    mol : Chem.Mol
        The molecule to validate.

    Returns
    -------
    validated_mol : Chem.Mol | None
        The sanitized molecule if valid, else None.
    """
    Chem.RemoveStereochemistry(mol)
    Chem.AssignStereochemistry(mol, cleanIt=True, force=True)
    Chem.AssignStereochemistryFrom3D(mol)

    for a in mol.GetAtoms():
        a.SetNoImplicit(True)  # type: ignore
        if a.HasProp("_MolFileHCount"):
            a.ClearProp("_MolFileHCount")

    flags = Chem.SanitizeFlags.SANITIZE_ALL & ~Chem.SanitizeFlags.SANITIZE_ADJUSTHS  # type: ignore

    # Full sanitization, minus ADJUSTHS
    err = Chem.SanitizeMol(
        mol,
        sanitizeOps=flags,
        catchErrors=True,
    )

    # Non-zero bitmask means some step failed
    if err:
        return None

    mol.UpdatePropertyCache(strict=True)
    return mol


def non_fragmented(mol: Chem.Mol) -> Optional[Chem.Mol]:
    """Return the original molecule if it is non-fragmented.

    Parameters
    ----------
    mol : Chem.Mol
        The molecule to check.

    Returns
    -------
    non_fragmented_mol : Chem.Mol | None
        The original molecule if it is fragmented, else None.
    """
    if len(Chem.GetMolFrags(mol)) == 1:
        return mol

    return None


def relax_geometry(mol: Chem.Mol) -> Chem.Mol | None:
    """Relax the geometry of a molecule using GFN-FF optimization.

    Parameters
    ----------
    mol : Chem.Mol
        The molecule to relax.

    Returns
    -------
    relaxed_mol : Chem.Mol | None
        The molecule with relaxed geometry. If any runtime errors, returns None.
    """
    with temporary_workdir():
        # Write molecule to XYZ file
        xyz_file = "input.xyz"
        output_file = "xtbopt.xyz"

        # Convert RDKit mol to XYZ format
        mol_block = Chem.MolToXYZBlock(mol)
        with open(xyz_file, "w") as f:
            f.write(mol_block)

        # Run xtb with GFN-FF optimization
        os.system(f"xtb {xyz_file} --opt --gfnff  > /dev/null 2>&1")

        if not os.path.exists(output_file):
            return None

        # Load optimized structure back into RDKit
        optimized_mol = Chem.MolFromXYZFile(output_file)  # type: ignore
        if optimized_mol is None:
            return None

        # Copy the optimized coordinates to the original molecule
        # to preserve bond orders, aromaticity, etc.
        conf = optimized_mol.GetConformer()
        original_conf = mol.GetConformer()
        for i in range(mol.GetNumAtoms()):
            pos = conf.GetAtomPosition(i)
            original_conf.SetAtomPosition(i, pos)  # type: ignore

    return mol



"""ImageReward from https://arxiv.org/abs/2304.05977."""

from typing import Any

import ImageReward as RewardModel  # type: ignore
import torch
from torchvision.transforms.functional import to_pil_image

from flowgym import FGTensor, Reward
from flowgym.registry import reward_registry


@reward_registry.register("images/image_reward")
class ImageReward(Reward[FGTensor]):
    """ImageReward that scores images based on a learned model of human preferences.

    Source: https://arxiv.org/abs/2304.05977
    """

    def __init__(self) -> None:
        self.reward_model = RewardModel.load("ImageReward-v1.0")

    def __call__(self, x: FGTensor, **kwargs: Any) -> tuple[torch.Tensor, torch.Tensor]:
        """Compute the image reward for a batch of images.

        Parameters
        ----------
        x : tensor, shape (B, C, H, W), values in [0, 1]
            A batch of images.

        Returns
        -------
        rewards : torch.Tensor, shape (B,)
            Image rewards.
        """
        if "prompt" not in kwargs:
            raise ValueError("ImageReward requires a 'prompt' keyword argument.")

        if x.min() < 0 or x.max() > 1:
            raise ValueError(f"`x` must have values in [0, 1], got [{x.min()}, {x.max()}]")

        prompt = kwargs["prompt"]

        with torch.autocast("cuda", enabled=False):
            pil_imgs = [to_pil_image(img.cpu().to(dtype=torch.float)) for img in x]

            rewards = []
            for prompt_txt, pil_img in zip(prompt, pil_imgs):
                rewards.append(self.reward_model.score(prompt_txt, pil_img))

        return torch.tensor(rewards), torch.ones(x.shape[0])



"""Utility classes and functions for Value Matching."""


class Report:
    """A simple class to keep track of statistics during training and logging them during training.

    It keeps track of every update and it keeps track of running statistics. The running statistics
    are reset at every log (call to __str__).
    """

    def __init__(self) -> None:
        self.stats: dict[str, list[float]] = {}
        self.running_means: dict[str, tuple[float, int]] = {}

    def update(self, **kwargs: float) -> None:
        """Update statistics with new values."""
        for key, value in kwargs.items():
            if key not in self.stats:
                self.stats[key] = []

            self.stats[key].append(value)

            running_mean, num_updates = self.running_means.get(key, (0, 0))
            self.running_means[key] = (running_mean + value, num_updates + 1)

    def __str__(self) -> str:
        strings = []
        for key, (running_mean, num_updates) in self.running_means.items():
            avg_value = running_mean / num_updates
            strings.append(f"{key}: {avg_value:.3f}")

        # Reset
        self.running_means = {}

        return ", ".join(strings)



"""Validity reward for molecules."""

from typing import Any, Optional

from rdkit import Chem
from rdkit.Chem import QED

from flowgym.molecules.rewards.base import MoleculeReward
from flowgym.registry import reward_registry


@reward_registry.register("molecules/qed")
class QEDReward(MoleculeReward):
    """QED reward for molecules."""

    def __init__(self, atom_type_map: Optional[list[str]] = None) -> None:
        super().__init__(atom_type_map, True, True, False)

    def compute_reward(self, mol: Chem.Mol, **kwargs: Any) -> float | None:
        """Compute Quantitative Estimate of Drug-likeness (QED) score."""
        return QED.qed(mol)  # type: ignore



"""Image reward functions for Flow Gym."""

from .aesthetic import AestheticReward
from .compression import CompressionReward, IncompressionReward
from .image_reward import ImageReward

__all__ = ["AestheticReward", "CompressionReward", "ImageReward", "IncompressionReward"]



"""Environments."""

from .base import Environment
from .endpoint import EndpointEnvironment
from .epsilon import EpsilonEnvironment
from .score import ScoreEnvironment
from .velocity import VelocityEnvironment

__all__ = [
    "EndpointEnvironment",
    "Environment",
    "EpsilonEnvironment",
    "ScoreEnvironment",
    "VelocityEnvironment",
]
